import 'package:dio/dio.dart';
import 'package:fe_network/fe_network.dart';
import 'package:fe_network/fe_network_entity.dart';
import 'package:fe_network/fe_network_error.dart';

typedef FeNetworkErrorCallBack = dynamic Function(FeNetworkError error);

enum FeNetworkMethod {
  get,
  put,
  post,
  delete,
}

class FeNetworkManager {

  Dio _dio;

  static final FeNetworkManager _instance = FeNetworkManager._internal();

  factory FeNetworkManager() => _instance;

  FeNetworkManager._internal() {
    if (_dio == null) {
      BaseOptions options = BaseOptions(
        contentType: Headers.jsonContentType,
        responseType: ResponseType.json,
        receiveDataWhenStatusError: false,
        connectTimeout: 30000,
        receiveTimeout: 120000,
      );
      _dio = Dio(options);
    }  
  }

  static FeNetworkManager shareInstance({String baseUrl}) {
    return _instance._baseUrl(baseUrl);
  }

  FeNetworkManager _baseUrl(String baseUrl) {
    if (_dio != null) {
      _dio.options.baseUrl = baseUrl;
    }
    return this;
  }

  Future<Map<String, dynamic>> request(
      FeNetworkMethod method,
      String path, {
        data,
        Map<String, dynamic> queryParameters,
        Options options,
        FeNetworkErrorCallBack onError,
      }) async {

    var resultOptions = (options == null ? Options() : options);
    resultOptions.method = _methodString(method);

    try {
      var response = await _dio.request(
        path,
        data: data,
        queryParameters: queryParameters,
        options: options,
      );

      /*
      * normal interface return structure
      * 1：
      * {
      * "code": "",
      * "message": "",
      * "data": {}
      * }
      * 
      * 2：
      *{
      * "code": "",
      * "message": "",
      * "data": []
      * }
      *
      * */
      if (response.data is Map<String, dynamic>) {
        final responseCode = response.data[keyCode];
        if (responseCode == keySuccessCode) {
          return response.data;
        }  else {
          var errMsg = response.data[keyMessage];
          onError(FeNetworkError(errCode: responseCode, errMsg: errMsg));
          return null;
        }
      } else {
        var errMsg;
        var errCode;
        if (response == null || response.data == null) {
          errMsg = 'Response data is null';
          errCode = '-110';
        }  else {
          errMsg = 'Response data returned is not Map<String, dynamic>';
          errCode = '-111';
        }
        onError(FeNetworkError(errCode: errCode, errMsg: errMsg));
        return null;
      }
    } on DioError catch (e) {
      if (onError != null) {
        onError(_handleDioError(e));
      }
      return null;
    }
  }

  FeNetworkError _handleDioError(DioError error) {
    var code;
    var msg;
    switch (error.type) {
      case DioErrorType.CONNECT_TIMEOUT: // NSURLErrorCannotConnectToHost
        code = '-1004';
        msg = 'Connect timeout';
        break;

      case DioErrorType.SEND_TIMEOUT: // NSURLErrorTimedOut
        code = '-1001';
        msg = 'Request timeout';
        break;

      case DioErrorType.RECEIVE_TIMEOUT: // NSURLErrorBadServerResponse
        code = '-1011';
        msg = 'Response timeout';
        break;

      case DioErrorType.RESPONSE:
        code = '${error.response.statusCode}';
        msg = error.response.statusMessage;
        break;

      case DioErrorType.CANCEL: // NSURLErrorCancelled
        code = '-999';
        msg = 'Request cancelled';
        break;

      case DioErrorType.DEFAULT: // NSURLErrorUnknown
        code = '-1';
        msg = error.message;
        break;
    }
    return FeNetworkError(errCode: code, errMsg: msg);
  }

  String _methodString(FeNetworkMethod method) {
    var result;
    switch (method) {
      case FeNetworkMethod.get:
        result = 'GET';
        break;
      case FeNetworkMethod.put:
        result = 'PUT';
        break;
      case FeNetworkMethod.post:
        result = 'POST';
        break;
      case FeNetworkMethod.delete:
        result = 'DELETE';
        break;
    }
    return result;
  }

  Future<BaseEntity> requestEntity<T>(
      FeNetworkMethod method,
      String path,
      T Function(Map<String, dynamic>) parseDataMap, {
        data,
        Map<String, dynamic> queryParameters,
        Options options,
      }) async {

    var resultOptions = (options == null ? Options() : options);
    resultOptions.method = _methodString(method);

    try {
      var response = await _dio.request(
        path,
        data: data,
        queryParameters: queryParameters,
        options: options,
      );

      /*
      * normal interface return structure
      * 1：
      * {
      * "code": "",
      * "message": "",
      * "data": {}
      * }
      *
      * 2：
      *{
      * "code": "",
      * "message": "",
      * "data": []
      * }
      *
      * */
      if (response.data is Map<String, dynamic> || response.data is List) {
        return BaseEntity.fromJson(response.data, parseDataMap);
      } else {
        if (response == null || response.data == null) {
          return BaseEntity(code: '-110', message: 'Response data is null', requestId: '', data: null, isSuccess: false);
        }  else {
          return BaseEntity(code: '-111', message: 'Response data returned is not Map<String, dynamic> or List structure', requestId: '', data: response.data, isSuccess: false);
        }
      }

    } on DioError catch (e) {
      return _handleError(e);
    }
  }

  BaseEntity _handleError(DioError error) {
    var code;
    var msg;
    switch (error.type) {
      case DioErrorType.CONNECT_TIMEOUT:
        code = '-1004';
        msg = 'Connect timeout';
        break;

      case DioErrorType.SEND_TIMEOUT:
        code = '-1001';
        msg = 'Request timeout';
        break;

      case DioErrorType.RECEIVE_TIMEOUT:
        code = '-1011';
        msg = 'Response timeout';
        break;

      case DioErrorType.RESPONSE:
        code = '${error.response.statusCode}';
        msg = error.response.statusMessage;
        break;

      case DioErrorType.CANCEL:
        code = '-999';
        msg = 'Request cancelled';
        break;

      case DioErrorType.DEFAULT:
        code = '-1';
        msg = error.message;
        break;
    }
    return BaseEntity(code: code, message: msg, requestId: '', data: null, isSuccess: false);
  }
}

